//===========================================================================
//
//  CRYSTALFONTZ CFAG128128A family Demo Code
//    https://www.crystalfontz.com/product/cfag128128atmitz
//    https://www.crystalfontz.com/product/cfag128128ayyhtz
//
//  The controller is a Toshiba T6963 / RAiO Technology RA6963
//  or compatible:
//    https://www.crystalfontz.com/controllers/Toshiba/T6963C/
//    https://www.crystalfontz.com/controllers/RAiOTechnology/RA6963/
//
//  Seeeduino v4.2, an open-source 3.3v capable Arduino clone.
//    https://www.seeedstudio.com/Seeeduino-V4.2-p-2517.html
//    https://github.com/SeeedDocument/SeeeduinoV4/raw/master/resources/Seeeduino_v4.2_sch.pdf
//===========================================================================
//
//  2019-08-22 Brent Crosby
//
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//============================================================================
// LCD control lines
// -------+-----+-----------------------------+--------+
//   ARD  |Port | Function       (Symbol,Pin) | Wire   |
// -------+-----+-----------------------------+--------+
//  #8/D8 | PB0 | Chip Enable Signal  (CE, 7) | Purple |
//  #9/D9 | PB1 | Reset            (RESET,10) | Blue   |
// #14/A0 | PC0 | Data/Instruction   (C/D, 8) | Yellow |
// #15/A1 | PC1 | Write              (/WR, 5) | Grey   |
// #16/A2 | PC2 | READ               (/RD, 6) | Brown  |
// #17/A3 | PC3 | Font Select         (FS,19) | White  |
// -------+-----+-----------------------------+--------+
//  #0/D0 | PD0 | LCD D10            (DB0,11) | Black  |
//  #1/D1 | PD1 | LCD D11            (DB1,12) | Brown  |
//  #2/D2 | PD2 | LCD D12            (DB2,13) | Red    |
//  #3/D3 | PD3 | LCD D13            (DB3,14) | Orange |
//  #4/D4 | PD4 | LCD D14            (DB4,15) | Yellow |
//  #5/D5 | PD5 | LCD D15            (DB5,16) | Green  |
//  #6/D6 | PD6 | LCD D16 Share: Tx  (DB6,17) | Blue   |
//  #7/D7 | PD7 | LCD D17 Share: Rx* (DB7,18) | Purple |
// -------+-----+-----------------------------+--------+
// *You will need to disconnect D7 during programming.
//
//  Note: there also needs to be a "contrast" potentiometer
//        (10K~20K) connected as follows:
//
//          One end:   5v             (red lead of meter)
//          Wiper:     LCD pin4 (Vo)  (black lead of meter)
//          Other end: LCD pin 9 (Vee) 
//
// Set this initially so the your meter reads ~16.3v
// -------------------------------------------------------------
#define CS_CLR_SELECT       (PORTB &= ~(0x01))
#define CS_SET_DESELET      (PORTB |=  (0x01))
#define RESET_CLR_RESET     (PORTB &= ~(0x02))
#define RESET_SET_OPERATE   (PORTB |=  (0x02))


#define RS_SET_COMMAND      (PORTC |=  (0x01))
#define RS_CLR_DATA         (PORTC &= ~(0x01))

#define WR_CLR_WRITE        (PORTC &= ~(0x02))
#define WR_SET_IDLE         (PORTC |=  (0x02))
#define RD_CLR_READ         (PORTC &= ~(0x04))
#define RD_SET_IDLE         (PORTC |=  (0x04))
#define FS_CLR_8x8          (PORTC &= ~(0x08))
#define FS_SET_6x8          (PORTC |=  (0x08))
#define RV_CLR_NORMAL       (PORTC &= ~(0x10))
#define RV_SET_INVERT       (PORTC |=  (0x10))
#define SET_LCD_PORT_OUT    (DDRD=0xFF)
#define SET_LCD_PORT_IN     (DDRD=0x00)
#define LCD_DATA_SET_OUT    PORTD
#define LCD_DATA_READ_IN    PIND
//============================================================================
// Depend on / enforce CS, RESET, WR, and RD idle / high / inactive.
//============================================================================
#include <Arduino.h>
#include "bitmaps.h"
//============================================================================
// ref: https://github.com/ct2034/T6963_Lib/blob/master/T6963.h

#define T6963_CURSOR_PATTERN_SELECT      (0xA0)
#define T6963_DISPLAY_MODE               (0x90)
#define T6963_MODE_SET                   (0x80)
#define T6963_SET_CURSOR_POINTER         (0x21)
#define T6963_SET_OFFSET_REGISTER        (0x22)
#define T6963_SET_ADDRESS_POINTER        (0x24)
#define T6963_SET_TEXT_HOME_ADDRESS      (0x40)
#define T6963_SET_TEXT_AREA              (0x41)
#define T6963_SET_GRAPHIC_HOME_ADDRESS   (0x42)
#define T6963_SET_GRAPHIC_AREA           (0x43)
#define T6963_SET_DATA_AUTO_WRITE        (0xB0)
#define T6963_SET_DATA_AUTO_READ         (0xB1)
#define T6963_AUTO_RESET                 (0xB2)

#define T6963_DATA_WRITE_AND_INCREMENT   (0xC0)
#define T6963_DATA_READ_AND_INCREMENT    (0xC1)
#define T6963_DATA_WRITE_AND_DECREMENT   (0xC2)
#define T6963_DATA_READ_AND_DECREMENT    (0xC3)
#define T6963_DATA_WRITE_AND_NONVARIALBE (0xC4)
#define T6963_DATA_READ_AND_NONVARIABLE  (0xC5)

#define T6963_SCREEN_PEEK                (0xE0)
#define T6963_SCREEN_COPY                (0xE8)

//============================================================================
void write_data_6963(uint8_t data)
  {
  //Apply the data to the port
  LCD_DATA_SET_OUT=data;
  //Select the chip
  CS_CLR_SELECT;
  //Select the data register
  RS_CLR_DATA;
  //Make the write pulse
  WR_CLR_WRITE;
  WR_SET_IDLE;
  _delay_us(10); //emprical
  //Deslect the chip
  CS_SET_DESELET;
  }
//============================================================================
void write_command_6963(uint8_t command)
  {
  //Apply the data to the port
  LCD_DATA_SET_OUT=command;
  //Select the chip
  CS_CLR_SELECT;
  //Select the command register
  RS_SET_COMMAND;
  //Make the write pulse
  WR_CLR_WRITE;
  WR_SET_IDLE;
  _delay_us(10); //emprical
  //Deslect the chip
  CS_SET_DESELET;
  }
//============================================================================
void LocateXY(uint8_t x,uint8_t y,uint8_t mode)
  {
  uint16_t
    address;
  address = LCD_Width_Bytes*y + x;
  // mode: GFX = 1, Text = 0
  if(mode)
    {
    address|=0x0800;
    }
  // Send the lower 8 bits
  write_data_6963(address);
  // Send the upper 8 bits
  write_data_6963(address>>8);
  // Then the command.
  write_command_6963(T6963_SET_ADDRESS_POINTER);  
  }
//----------------------------------------------------------------------------
void Fill_Checkerboard() 
  {
  uint8_t
    i;
  uint8_t
    j;
  uint8_t
    data;
  data=0xAA;
  //Move to the correct location, graphic
  LocateXY(0,0,1);
  // Turn on auto address increment
  write_command_6963(T6963_SET_DATA_AUTO_WRITE);
  for(j=0;j<LCD_Height_Dots;j++)
    {
    for(i=0;i<LCD_Width_Bytes;i++)
      {
      write_data_6963(data);
      }
    //Invert for the next line.
    data^=0xFF;
    }
  // Turn off auto address increment
  write_command_6963(T6963_AUTO_RESET);
  }
//----------------------------------------------------------------------------
void Fill_graphic_with_data( uint8_t FillByte )
  {
  uint16_t
    i;
  //Calculate nuber of bytes to write.
  i=LCD_Width_Bytes*LCD_Height_Dots;      

  //Move to the correct location, graphic
  LocateXY(0,0,1);
  // Turn on auto address increment
  write_command_6963(T6963_SET_DATA_AUTO_WRITE);
  for(i=0;i<LCD_Width_Bytes*LCD_Height_Dots;i++)
    {
    write_data_6963(FillByte);
    }
  // Turn off auto address increment
  write_command_6963(T6963_AUTO_RESET);
  }
//----------------------------------------------------------------------------
void Fill_text_with_data( uint8_t FillByte )
  {
  uint16_t
    i;
  //Calculate nuber of bytes to write.
  i=LCD_Width_Bytes*LCD_Height_Dots;      

  //Move to the correct location, graphic
  LocateXY(0,0,0);
  // Turn on auto address increment
  write_command_6963(T6963_SET_DATA_AUTO_WRITE);
  for(i=0;i<LCD_Width_Bytes*LCD_Height_Dots/8;i++)
    {
    write_data_6963(FillByte);
    }
  // Turn off auto address increment
  write_command_6963(T6963_AUTO_RESET);
  }
//============================================================================
void Show_Bitmap(const uint8_t *Bitmap_Data)
  {
  uint8_t
    i;
  uint8_t
    j;
  //Move to the correct location, graphic
  LocateXY(0,0,1);
  // Turn on auto address increment
  write_command_6963(T6963_SET_DATA_AUTO_WRITE);

  //Sweeping down the lines
  for(j=0;j<LCD_Height_Dots;j++)
    {
    //Moving from left to right across each line
    for(i=0;i<LCD_Width_Dots/8;i++)
      {
      //Get the data from flash, send it to the controller.
      write_data_6963(pgm_read_byte(Bitmap_Data));
      //Move the pointer to the next byte
      Bitmap_Data++;
      }
    }
  // Turn off auto address increment
  write_command_6963(T6963_AUTO_RESET);
  }
//============================================================================

void Initialize_CFAG240128L()
  {
  //Clobber the display controller into submission. 
   RESET_CLR_RESET;
   _delay_ms(10);
   RESET_SET_OPERATE;
   _delay_ms(50);

  //Format is:
  //  Low Byte
  //  High Byte
  //  Command

  // set text home address
  write_data_6963(0x00);
  write_data_6963(0x00);
  write_command_6963(T6963_SET_TEXT_HOME_ADDRESS);  

  // set text area
  write_data_6963(LCD_Width_Bytes);
  write_data_6963(0x00); 
  write_command_6963(T6963_SET_TEXT_AREA);  

  // set graphic home address
  write_data_6963(0x00);
  write_data_6963(0x08);     // address 0x0800
  write_command_6963(T6963_SET_GRAPHIC_HOME_ADDRESS);  

  // set graphic area
  write_data_6963(LCD_Width_Bytes);
  write_data_6963(0x00);
  write_command_6963(T6963_SET_GRAPHIC_AREA);  

  // display mode - graphic on, text on
  write_command_6963(T6963_CURSOR_PATTERN_SELECT);  // 1 line cursor
  write_command_6963(T6963_MODE_SET | 0x01);
  write_command_6963(T6963_DISPLAY_MODE | 0x0C);  

  //Clear the graphic memory
  Fill_graphic_with_data(0x00);

  //Clear the text memory
  Fill_text_with_data(0);

  // Set the location of next data write
  write_data_6963(0x00);
  write_data_6963(0x00);
  write_command_6963(T6963_SET_ADDRESS_POINTER);  

  // Set the cursor position
  write_data_6963(0x00);
  write_data_6963(0x00);
  write_command_6963(T6963_SET_CURSOR_POINTER);  
  }
//============================================================================
void setup()
  {
  //Initialize / idle the ports & start reset pulse.
  CS_SET_DESELET;
  RESET_CLR_RESET;
  RS_SET_COMMAND;
  WR_SET_IDLE;
  RD_SET_IDLE;
  FS_CLR_8x8;
  RV_CLR_NORMAL;
  LCD_DATA_SET_OUT=0x00;
    
  //Set port directions
  DDRB = 0x03;
  DDRC = 0x1F;
  SET_LCD_PORT_OUT;
  
  //Set up the 6963 controller
  Initialize_CFAG240128L();
  }
//============================================================================
void loop()
  {  
  Fill_Checkerboard();
  _delay_ms(2000);
    
  Show_Bitmap(&CFAG128128A_Logo[0][0]);
  _delay_ms(2000);

  Show_Bitmap(&Sophie128128[0][0]);
  _delay_ms(2000);

  Fill_graphic_with_data(0x00);
  _delay_ms(500);
  Fill_graphic_with_data(0x01);
  _delay_ms(500);
  Fill_graphic_with_data(0x03);
  _delay_ms(500);
  Fill_graphic_with_data(0x07);
  _delay_ms(500);
  Fill_graphic_with_data(0x0F);
  _delay_ms(500);
  Fill_graphic_with_data(0x1F);
  _delay_ms(500);
  Fill_graphic_with_data(0x3F);
  _delay_ms(500);
  Fill_graphic_with_data(0x7F);
  _delay_ms(500);
  Fill_graphic_with_data(0xFF);
  _delay_ms(500);

  Fill_graphic_with_data(0xFE);
  _delay_ms(500);
  Fill_graphic_with_data(0xFC);
  _delay_ms(500);
  Fill_graphic_with_data(0xF8);
  _delay_ms(500);
  Fill_graphic_with_data(0xF0);
  _delay_ms(500);
  Fill_graphic_with_data(0xE0);
  _delay_ms(500);
  Fill_graphic_with_data(0xC0);
  _delay_ms(500);
  Fill_graphic_with_data(0x80);
  _delay_ms(500);
  Fill_graphic_with_data(0x00);
  _delay_ms(1000);
  }
//============================================================================
